grammar Formules;

options {
//  backtrack=true;
//  memoize=true;
  language = Java;
}
@header {
package fr.cnav.saturne.dsl.parser;  
import fr.cnav.saturne.dsl.formules.*;
import fr.cnav.saturne.dsl.formules.FormulesGenerator;
import java.util.ArrayList;
import java.util.List; 
} 
@lexer::header {
package fr.cnav.saturne.dsl.parser;
}
@lexer::members{
  public void displayRecognitionError(String[] tokenNames,
                    RecognitionException e)
  {
    String hdr = getErrorHeader(e);
    String msg = getErrorMessage(e, tokenNames);
    emitErrorMessage("      [ANTLR Lexical Error] " +hdr+" "+msg);
  }
  
  public void recover(RecognitionException re) 
  {
    reportError(re);
    throw new RuntimeException(re);
  }
}
@members {

public String getCompleteMessage(RecognitionException e){
    String hdr = getErrorHeader(e);
    String msg = getErrorMessage(e, this.getTokenNames());
return  hdr + " " + msg;
}


private FormulesGenerator gen = new FormulesGenerator();
protected Object recoverFromMismatchedToken(IntStream input, int ttype, BitSet follow) throws RecognitionException {
   throw new MismatchedTokenException(ttype,input);
}
public Object recoverFromMismatchedSet(IntStream input,RecognitionException e,BitSet follow) throws RecognitionException {
  throw e;
}

//private String errorMessage = null;
//
public void displayRecognitionError(String[] tokenNames, RecognitionException e) {
    String hdr = getErrorHeader(e);
    String msg = getErrorMessage(e, tokenNames);
    emitErrorMessage("      [ANTLR Syntax Error] " +hdr+" "+msg);
}
  private void checkDate(String date) throws RecognitionException {
    if (date.length() < 9) {
      throw new RecognitionException();
    }
    
    boolean correct = true;    
    String day = date.substring(0, 1);
    String mont = date.substring(2, 3);
    String year = date.substring(4, 7);
    try {
      int dayI = Integer.parseInt(day);
      if (dayI > 31 || dayI < 0) {
        correct = false;
      }
      int monthI = Integer.parseInt(mont);
      if (monthI > 12 || monthI < 0) {
        correct = false;
      }
      int yearI = Integer.parseInt(year);
      if (yearI < 0) {
        correct = false;
      }
    } catch (Exception e) {
      correct = false;
    }
    if(!correct){
       throw new RecognitionException();
    }
  }


}
@rulecatch {
catch (RecognitionException e) {
    reportError(e);
    throw e;
}
}

entry returns [Expression result] : f=formule EOF {$result=$f.result;}
;

// Cette règle sert juste à identifier le point d'entrée de manière univoque.
formule returns [Expression result] 
@init{
  $result = gen.error("Error on parsing formule expression" );
}
: exp1=logicalExpr     {$result=$exp1.result;}
;

letExpr returns [Expression result]
 : 'let' {$result= gen.let();} b0=binding{((Let)$result).getBindings().add($b0.result);} (','b1=binding{((Let)$result).getBindings().add($b1.result);})* 'inexpr' body=formule {((Let)$result).setIn($body.result);}
;
binding returns [Binding result]: FIELD ':=' exp=formule {$result = gen.binding($FIELD.text,$exp.result);} 
; 
quantifiedExpr returns [Expression result] : q=QUANTIFIER vars=variableDecl 'satisfies' body=formule{
$result = gen.quantified($q.text,$body.result,$vars.vars);
} 
;
variableDecl returns [List<IteratorVariable> vars] 
@init{
  $vars = new ArrayList<IteratorVariable>();
} :
elt1=element {$vars.add(gen.iteratorVariable("self",(FieldRef)$elt1.result));}
| (id1=FIELD ':' elt1=element {$vars.add(gen.iteratorVariable($id1.text,(FieldRef)$elt1.result));}) (',' id2=FIELD ':' elt2=element{$vars.add(gen.iteratorVariable($id2.text,(FieldRef)$elt2.result));})*
;


//Grammaire des expressions 
// A noter que le traitement du NOT était ambigüe dans la BNF fourni:
// il pouvait être associé de différentes manière. "not A and B" pouvait être interprété "not (A and B)" 
// ou (not A) and B. On a choisi la seconde solution dans cette grammaire. A voir ce qui convient au final.

// La grammaire des expressions est dé-récursifiée par rapport à celle du CDC mais le langage reconnu est le même.
logicalExpr returns [Expression result] 
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing logical expression on column : " + position);
}
: leftOp=orExpr {$result=$leftOp.result;} (CONNECTOR rightOp=logicalExpr
{
  if("=>".equals($CONNECTOR.text)){
    $result = gen.imply($leftOp.result,$rightOp.result);
  } else {
    $result = gen.equiv($leftOp.result,$rightOp.result);
  }
})?
; 
orExpr returns [Expression result]
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing or expression on column : " + position);
}
: leftOp=andExpr {$result=$leftOp.result;} ('or' rightOp=orExpr   {$result = gen.or (gen.expressions($leftOp.result,$rightOp.result));})?
;
andExpr returns [Expression result] 
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing and expression on column : " + position);
}
: leftOp=notExpr {$result=$leftOp.result;} ('and' rightOp=andExpr {$result = gen.and(gen.expressions($leftOp.result,$rightOp.result));})?
;
notExpr returns [Expression result] 
@init{
  $result = gen.error("Error on parsing no expression on column" );
}
: ('not'{$result=gen.not(null);} )? operand=comparisonExpr
{
  if($result != null && result instanceof Not){
    ((Not)$result).setOperand($operand.result);
  } else {
    $result = $operand.result;
  } 
} 
;
comparisonExpr returns [Expression result]
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing compare expression on column : " + position);
}
: leftOp=arithmeticExpr {$result=$leftOp.result;} (COMPOP rightOp=arithmeticExpr
{
if ("<".equals($COMPOP.text)){
    $result = gen.lower($leftOp.result,$rightOp.result);
}else if (">".equals($COMPOP.text)){
    $result = gen.greater($leftOp.result,$rightOp.result);
}else if ("<=".equals($COMPOP.text)){
    $result = gen.lowerEqual($leftOp.result,$rightOp.result);
}else if (">=".equals($COMPOP.text)){
    $result = gen.greaterEqual($leftOp.result,$rightOp.result);
}else if ("=".equals($COMPOP.text)){
    $result = gen.equals($leftOp.result,$rightOp.result);
}else if ("!=".equals($COMPOP.text)){
    $result = gen.diff($leftOp.result,$rightOp.result);
}else if ("in".equals($COMPOP.text)){
  if($rightOp.result instanceof ValueList){
    $result = gen.in($leftOp.result,(ValueList)$rightOp.result);
  } else {
   $result = gen.error("[Error] Parser: right operand of 'in' operator must be a value list. " + $rightOp.result.getClass().getName() + " found instead.");
  }
}
}
)?
;
arithmeticExpr returns [Expression result] 
@init{
  Token firstToken = $start;
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing arithmetic expression on column : " + position);
}
: leftOp=multExpr {$result=$leftOp.result;}(ADDOP rightOp=arithmeticExpr
{
  if("+".equals($ADDOP.text)){
    $result = gen.add(gen.expressions());
    ((Nary)$result).getOperands().add($leftOp.result);
    ((Nary)$result).getOperands().add($rightOp.result);
  } else {
    $result = gen.sub(gen.expressions());
    ((Nary)$result).getOperands().add($leftOp.result);
    ((Nary)$result).getOperands().add($rightOp.result);
  }
})?
;
multExpr returns [Expression result] 
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing mult expression on column : " + position);
}
: leftOp=atom {$result=$leftOp.result;}(MULTOP rightOp=multExpr
{
  if("*".equals($MULTOP.text)){
    $result = gen.mult(gen.expressions());
    ((Nary)$result).getOperands().add($leftOp.result);
    ((Nary)$result).getOperands().add($rightOp.result);
  } else {
    $result = gen.div(gen.expressions());
    ((Nary)$result).getOperands().add($leftOp.result);
    ((Nary)$result).getOperands().add($rightOp.result);
  }
})?
;
atom returns [Expression result] 
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing atom expression on column : " + position);
}
: val=literal  {$result=$val.result;} 
| elt=element  {$result=$elt.result;}
| call=functionCall {$result=$call.result;}
| '(' exp=formule ')' {$result=$exp.result;}
| listVal=listeValeur {$result=$listVal.result;}
| ref = varRef {$result=$ref.result;} 
| exp2=letExpr         {$result=$exp2.result;}
| exp3=quantifiedExpr  {$result=$exp3.result;}
;
// appels des fonctions définies dans le fichier de properties (e.g. #cle)
functionCall returns [Expression result]
@init{
 $result = gen.error("Error on parsing functionCall expression." );
 List<Expression> argsList = new ArrayList<Expression>();
 String functionName = null;
}
@after {
 $result=gen.functionCall(functionName);
 ((Call)$result).getArguments().addAll(argsList);
}
:  name=FIELD  {functionName = $name.text;}'('(arg1=formule{argsList.add($arg1.result);}(','arg2=formule{argsList.add($arg2.result);})* )? ')'
;
// appels des fonctions dites 'builtin' 
//builtinCall returns [Expression result]
//@init{
// $result = gen.error("Error on parsing builtinCall expression on column : mettre la variable de column FIXMEFIXME -- OHA" );
// List<Expression> argsList = new ArrayList<Expression>();
// fr.cnav.n4ds.formules.BuiltIn builtIn = null;
//}
//@after {
// $result=gen.builtInCall(builtIn);
// ((BuiltInCall)$result).getArguments().addAll(argsList);
//}
//:  
//name=builtIn {builtIn=$name.result;} '('(arg1=arithmeticExpr{argsList.add($arg1.result);}(','arg2=arithmeticExpr{argsList.add($arg2.result);})* )? ')'
//;
//builtIn returns [BuiltIn result]: '$add_jour'        {$result=gen.addJour();} 
//                                | '$add_mois'        {$result=gen.addMois();} 
//                                | '$add_annee'       {$result=gen.addAnnee();} 
//                                | '$sub_jour'        {$result=gen.subJour();} 
//                                | '$sub_mois'        {$result=gen.subMois();} 
//                                | '$sub_annee'       {$result=gen.subAnnee();} 
//                                | '$sub_string'      {$result=gen.subString();} 
//                                | '$period_included' {$result=gen.periodIncluded();} 
//                                | '$period_overlap'  {$result=gen.periodOverlap();}
//                                | '$annee'           {$result=gen.annee();} 
//                                | '$is_present' {$result=gen.isPresent();}
//                                | '$is_absent'{$result=gen.isAbsent();}
//                                | '$is_true'  {$result=gen.isTrue();}
//                                | '$starts_with'     {$result=gen.startsWith();} 
//;
//rubriques, et accès aux variables.
//element returns [Expression result]
//@init{
//  Token firstToken = $start; 
//  int position = firstToken.getCharPositionInLine();
//  $result = gen.error("Error on parsing element expression on column : " + position );
//}
//: rub=structure {$result = $rub.result;}| rub=groupe {$result = $rub.result;}| rub=sousGroupe {$result = $rub.result;}| rub=rubrique  {$result = $rub.result;} 
//;
varRef returns [Expression result]: '$'FIELD {$result = gen.varRef($FIELD.text);} 
;
contextSpec returns [String result]: '$'FIELD ':' {$result=$FIELD.text;}
;
element returns [Expression result]
@init{
  Token firstToken = $start; 
  int position = firstToken.getCharPositionInLine();
  $result = gen.error("Error on parsing element expression on column : " + position );
 String var = "root";
}
: (ctxt=contextSpec {var=$ctxt.result;})? rub=fieldRef {$result = $rub.result; ((FieldRef)$result).setContextVariable(var);}
;
fieldRef returns [FieldRef result]: FIELD
{
$result=gen.buildFieldRef($FIELD.text);
}
;
listeValeur returns [ValueList result] 
@init{ 
 List<Literal> valueList = new ArrayList<Literal>();
}
@after{
  $result=gen.valueList();
  $result.getValues().addAll(valueList);
}:
'{' lit1=literal {valueList.add($lit1.result);}(','lit2=literal {valueList.add($lit2.result);})* '}'
;

// Pour le literal ALPHANUM, nous retirons les simples cotes en debut et fin de valeur
literal returns [Literal result]:
    NUM { $result = gen.num($NUM.text);}
  | ALPHANUM {$result = gen.alphaNum($ALPHANUM.text.substring(1, $ALPHANUM.text.length()-1));}
  | DATE {$result = gen.date($DATE.text); checkDate($DATE.text); }
;

//Les structures/groupes/rubriques sont traités par des règles terminales.
//Les exemples comportait des sous-
//rubriques et sous rubriques dont 
//les numéros étaient codés alternativement sur deux ou trois chiffre.
//On prend en compte cette possibilité.
WS  : (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=HIDDEN;}
    ;
QUANTIFIER : 'some' | 'every' | 'none_or_every' | 'none'
;

DATE:  DIGIT DIGIT DIGIT DIGIT DIGIT DIGIT DIGIT DIGIT  ('D'|'d')
;
NUM:  DIGIT+ ('.' DIGIT DIGIT)?
;

CONNECTOR: '=>' | '<=>'
;
MULTOP : '*' | '/'
;
ADDOP : '+' | '-'
;
COMPOP : '=' | '<' | '>' | '<=' | '>=' | '!=' | 'in' 
;
ALPHANUM:  '\'' (~('\t'|'\u000C'|'\''|'\n'|'\r'))* '\''
; 
//ID: IDSTART (IDPART)*
//;
  
// TODO lettres + chiffres + _ + . + -
FIELD :  (IDSTART) (FIELDPART)*
;
////Order if ID and LIT is important since ID matches a subset of LIT.
//LIT: IDPART*
//;
fragment IDSTART: 'a'..'z' |  'A'..'Z' | '_';
fragment IDPART: 'a'..'z' |  'A'..'Z' | '_' | '-'|'0'..'9';
fragment FIELDPART: 'a'..'z' |  'A'..'Z' | '_' | '-'| '.'|'0'..'9';

fragment DIGIT : '0'..'9'
; 
